/* cross-section measurements
   #ifdef XSECTION only:
   is #included from sfdx.c TWICE:
  XSECTION_M=1: original molecule-based version
  XSECTION_M=0: whole configuration version
*/

#if XSECTION_M
double mXsection
#else
double cXsection
#endif
  (                                                       /*** [mc]Xsection */
   double grid,int mode,double RvdW,double Rscale,
   int freq, int maxs, int sizelimit,int *maxsize)
/*
  Calculates cross sections based on spheres with radii given by the
    van der Waals radii + probe radius RvdW, multiplied by Rscale
  grid = number of sample points per Angstrom
  freq : see the manual
  mode = 0: 32-point Gaussian integration
            (32 Gaussian points are first rotated by a random rot matrix;
             -- actually only 16 of them are used because of symmetry)
  mode = 1: one random direction
  mode = 2: (x+y),(x-y),z separately
  mode = 3: x,y,z separately
  mode = 4: average over 3 orthogonal axes (rotated by a random rot matrix)
  NOTE:
    modes 0,1,4 give the angle-averaged cross-section
    modes 2,3 give cross-sections in specified directions: suitable for
          recognizing orientational order, e.g. use mode=2 to see whether
          shear stress orders elongated molecules
  return value: sum of all cross sections; for modes 2,3 the 1st of three
    directions only
  sizelimit: max size of the grid array
  maxsize : max size actually encountered
*/
{
  static int modes[5]={16,1,3,3,3};
  static double weights[5]={0,1,1,1,1./3};
  static int warned=0,Maxsize=0;
  int overflow;

  molecule_t *mn;
  siteinfo_t *si;
  vector *r;

  int imode,ns,sp,n,i,dimx,dimy,fromx,tox,fromy,toy,ix;
  real rr,dy,minx,maxx,miny,maxy,w;

  unsigned summesh;
  unsigned char *mesh;

  vector ray,rvec,nx,ny;
  typedef real vector2[2];
  vector2 *proj;
#if XSECTION_M
  vector *sum;
  int *sumn;
  double Sum=0;
#else
  vector sum={0,0,0};
#endif
  int is,ifreq;
  char *s;

  static matrix sheardir={
    {0.70710678118654757, 0.70710678118654757, 0},
    {0.70710678118654757,-0.70710678118654757, 0},
    {0,                   0,                   1} };
  static matrix xyzdir={
    {1,0,0},
    {0,1,0},
    {0,0,1}};

  matrix R;

  if (mode<0 || mode>4) ERROR(("[mc]Xsection: mode=%d)",mode))

  if (freq<0) freq=-freq; else freq=1;

#if XSECTION_M
  if (maxs<0) maxs=nspec;
  Min(maxs,nspec)
  allocarrayzero(sum,maxs);
  allocarrayzero(sumn,maxs);

  loop (n,0,No.N) {
    mn=molec+n;
    ns=mn->ns;
    if (ns<2) continue; /* ignore monoatomics */
    sp=mn->sp;
    if (sp>=maxs) continue; /* ignore species */
    si=spec[sp]->si;
    r=rof(mn,cfg[0]->rp);
    allocarray(proj,ns);
#else
    if (maxs<0) maxs=No.s;
    Min(maxs,No.s)
    allocarray(proj,No.s);
#endif

    loop (ifreq,0,freq) {

      w=weights[mode]; /* overwritten for mode=0 */

      loop (imode,0,modes[mode]) {

        if (mode==0 || mode==4) rndmat(R);

        switch (mode) {
          /* only each second vector of sphint is taken because of symmetry */
          case 0: mpl(ray,R,sphint[imode*2].n); w=sphint[imode*2].w*2; break;
          case 1: rndonsph(ray); break;
          case 2: VV(ray,=sheardir[imode]) break;
          case 3: VV(ray,=xyzdir[imode]) break;
          case 4: VV(ray,=R[imode]); break; }
        /* now, ray is the direction of the projection */

        do rndsph(rvec); while (fabs(SCAL(rvec,ray))<0.0625);
        VECT(ny,ray,rvec)
        rr=sqrt(SQR(ny));
        VO(ny,/=rr)
        VECT(nx,ray,ny)
        /* now, nx and ny are axes in the projection plane, randomly rotated */

        /* PASS 1: determinimg min/max sizes of the mesh */
        maxx=maxy=-3e33;
        minx=miny=3e33;

#if XSECTION_M
        loop (i,0,ns) {
          is=i;
#else
        is=0;
        loop (n,0,No.N) {
          mn=molec+n;
          ns=mn->ns;
          sp=mn->sp;
          si=spec[sp]->si;
          r=rof(mn,cfg[0]->rp);
          loop (i,0,ns) {
#endif
            rr=(sitedef[si[i].st].LJ[0].RvdW+RvdW)*Rscale;

            proj[is][0]=SCAL(nx,r[i]);
            proj[is][1]=SCAL(ny,r[i]);

            Max(maxx,proj[is][0]+rr)
            Max(maxy,proj[is][1]+rr)
            Min(minx,proj[is][0]-rr)
            Min(miny,proj[is][1]-rr)
#if !XSECTION_M
            is++; } }
        if (is!=No.s) ERROR(("internal: is!=No.s"))
#else
        }
#endif

        dimx=grid*(maxx-minx)+1.0001;
        dimy=grid*(maxy-miny)+1.0001;

        /* to prevent interference errors */
        minx-=rnd()/grid;
        miny-=rnd()/grid;

/*.....put3(minx,maxx,dimx)*/
/*.....put3(miny,maxy,dimy)*/

        overflow=dimx*dimy>sizelimit;
        Max(Maxsize,dimx*dimy) *maxsize=Maxsize;
        if (overflow && warned<=4) {
          WARNING(("Xsec: grid array %dx%d too large - truncated, Xsec underestimated%s",dimx,dimy,
                   warned==4?"\n(more occurrences not reported)":""))
          /* not clear what's the best strategy if grid not enough */
          /* this one tries centering, good for some atoms trying to escape */

          do {
            double M=fmax(fmax(maxx,maxy),fmax(-minx,-miny));

            if (maxx/M>0.9) maxx*=.99;
            if (maxy/M>0.9) maxy*=.99;
            if (-minx/M>0.9) minx*=.99;
            if (-miny/M>0.9) miny*=.99;
            dimx=grid*(maxx-minx);
            dimy=grid*(maxy-miny);
          } while (dimx*dimy>sizelimit);

/*.....prt("%g %g  %g %g",minx,maxx,miny,maxy);*/
          warned++; }

        /* CREATE MESH
           - indices of the mesh array are (ix,iy), 0<=ix<dimx, 0<=iy<dimy
           - they represent grid points ((ix+1)/grid, (iy+1)/grid)
           - the map function is mesh[ix*dimy+iy] */
        alloczero(mesh,dimx*dimy); /* 2D array, my own map function */

        /* PASS 2: projecting particles on the mesh */
#if XSECTION_M
        loop (i,0,ns) {
          is=i;
#else
        is=0;
        loop (n,0,No.N) {
          mn=molec+n;
          ns=mn->ns;
          sp=mn->sp;
          si=spec[sp]->si;
          r=rof(mn,cfg[0]->rp);
          loop (i,0,ns) {
#endif
            rr=(sitedef[si[i].st].LJ[0].RvdW+RvdW)*Rscale;
            proj[is][0]-=minx;
            proj[is][1]-=miny;
            fromx=(int)(grid*(proj[is][0]-rr));
            tox  =(int)(grid*(proj[is][0]+rr));
            if (fromx<0 || tox>dimx) {
              fromx=tox=0;
              if (!overflow) ERROR(("Xsec: xrange=%d %d",fromx,tox)) }

            rr*=rr*Sqr(grid);
            loop (ix,fromx,tox) {
              dy=rr-Sqr(ix+1-proj[is][0]*grid);
              if (dy>=0) {
                dy=sqrt(dy);
                fromy=(int)(grid*proj[is][1]-dy);
                toy  =(int)(grid*proj[is][1]+dy);
                if (fromy<0 || toy>dimy) {
                  fromy=toy=0;
                  if (!overflow) ERROR(("Xsec: yrange=%d %d",fromy,toy)) }
                if (toy-fromy) memset(mesh+ix*dimy+fromy,1,toy-fromy); } }
#if !XSECTION_M
          is++; } }
        if (is!=No.s) ERROR(("internal: is!=No.s"))
#else
        }
#endif

      summesh=0;
      loop (i,0,dimx*dimy) summesh+=mesh[i];

#if 0
      {
        /* DEBUG */
        static int count;
        int i,j;
#  define SX 10
#  define SY 10
        loop (j,0,dimy/SY) {
          _n
            loop (i,0,dimx/SX)
            if (mesh[i*SY*dimy+j*SX]) prtc('%'); else prtc('.'); }
        _n
          if (count++>20) exit(0);
      }
#endif /*# 0 */

#if !XSECTION_M

      if (mode==2 || mode==3) sum[imode]+=summesh*w;
      else sum[0]+=summesh*w;
#else
      if (mode==2 || mode==3) sum[sp][imode]+=summesh*w;
      else sum[sp][0]+=summesh*w;
#endif

      free(mesh); } /* imode */
    } /* ifreq */

#if !XSECTION_M
  if (mode==2 || mode==3)
    loop (i,0,3) {
      rr=sum[i]/((double)freq*Sqr(grid));
      s=string("%s Xsection[cfg]",
               i*4+(mode==2 ? "x+y\0x-y\0z" : "x\0HAy\0HAz"));
      StaAdd(s,rr);
      if (0) prt_("%s=%g  %c",s,rr,"\0\n"[i==2]); }
  else {
    s="Xsection[cfg]";
    rr=sum[0]/(freq*Sqr(grid));
    StaAdd(s,rr);
    if (0) prt("%s=%g",s,rr); }

  free(proj);

  return rr;

#else
    sumn[sp]++;
    free(proj); } /* n */

  loop (sp,0,maxs) if (sumn[sp]) {
    Sum+=sum[sp][0];
    if (mode==2 || mode==3)
      loop (i,0,3) {
      rr=sum[sp][i]/(freq*sumn[sp]*Sqr(grid));
      s=string("%s Xsection[%d]",
               i*4+(mode==2 ? "x+y\0x-y\0z" : "x\0HAy\0HAz"),sp);
      StaAdd(s,rr);
      if (0) prt_("%s=%g  %c",s,rr,"\0\n"[i==2]); }
    else {
      s=string("Xsection[%d]",sp);
      rr=sum[sp][0]/(freq*sumn[sp]*Sqr(grid));
      StaAdd(s,rr);
      if (0) prt("%s=%g",s,rr); } }

  free(sumn);
  free(sum);

  return Sum/Sqr(grid);

#endif
}
